#!/usr/bin/env bash

dptools input dft_inputs/0?_???.db
#dptools input dft_inputs/0*.traj
#dptools input dft_inputs/0* # should fail
#dptools input -n 200 dft_inputs/0*.traj
#dptools input -p path_test dft_inputs/0*.traj
